# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import itertools
import GridUtils
import PGUtils
numpy = GridUtils.import_numpy()
import random

class GameState (object):
    # if dimensions is not specified, "grid" property should be set manually
    def __init__(self, dimensions=None, params=None, players=2, reset=True):
        if dimensions: self.grid = numpy.zeros(dimensions, int)
        self.game_params = params
        self.num_players = players
        self.current_player = 1
        self.extra_state = None
        if reset: self.reset()
        
    def serialize(self):
        return {'shape': self.grid.shape, 'values': list(self.grid.flat), 'extra_state': self.extra_state,
                'num_players': self.num_players, 'current_player': self.current_player}

    # array methods
    def dimensions(self):
        return self.grid.shape
        
    def valid_coords(self, c):
        return GridUtils.valid_coords(self.grid, c)
        
    def index_array(self):
        return GridUtils.index_array(self.grid.shape)
        
    def itervalues(self):
        return self.grid.flat
        
    def itercoords(self):
        return numpy.ndindex(self.grid.shape)
        
    def enumerate_all(self):
        return numpy.ndenumerate(self.grid)
        
    def neighbor_coords(self, center, dist=1):
        return GridUtils.neighbor_coords(self.grid, center, dist)
        
    def enumerate_neighbors(self, center, dist=1):
        return GridUtils.enumerate_neighbors(self.grid, center, dist)
        
    def enumerate_coords_with_value(self, val):
        return GridUtils.enumerate_coords_with_value(self.grid, val)
            
    def enumerate_neighbor_coords_with_value(self, center, val, dist=1):
        return GridUtils.enumerate_neighbor_coords_with_value(self.grid, center, val, dist)
            
    def enumerate_values_with_direction(self, start, direction):
        return GridUtils.enumerate_values_with_direction(self.grid, start, direction)        
    # end array methods
    
    def next_player(self):
        if self.current_player>=self.num_players: 
            return 1
        return self.current_player+1
        
    def previous_player(self):
        if self.current_player==1:
            return self.num_players
        return self.current_player-1
    
    # returns a new game state after the current player's move results in the given grid
    # the grid argument is not copied, so modifications will affect the grid in the returned state
    # if subclasses need to copy additional state, they should override and call superclass first
    # (overriding this shouldn't normally be needed since subclasses can override copy_extra_state, see below)
    def next_state_with_grid(self, grid):
        # pass reset=False because we don't need to initialize the grid in __init__
        c = type(self)(None, self.game_params, self.num_players, reset=False)
        c.grid = grid
        c.current_player = self.next_player()
        if self.extra_state is not None:
            c.extra_state = self.copy_extra_state()
        return c
        
    # Called to copy the extra_state attribute. Default implementation does a shallow dict copy, subclasses
    # should override for custom behavior (for example, making deep copies of list values)
    def copy_extra_state(self):
        return self.extra_state.copy()
        
    # returns True if game is over. Calls successors() and returns false if it sees a valid move
    # subclasses can override if there's a faster way
    def is_game_over(self):
        return not any(self.successors())
        
    # sets game to start condition
    def reset(self):
        self.current_player = 1
        self.init_grid_from_params()
        
    # default looks for "dimensions" and "positions" in params dict
    def init_grid_from_params(self):
        if self.game_params and self.game_params.get('dimensions'):
            self.grid = numpy.zeros(self.game_params.get('dimensions'), int)
            for pnum in self.game_params.get('positions', {}):
                for pos in self.game_params['positions'][pnum]:
                    # pnum needs to be a string so it can be sent over JSON as a dictionary key
                    self.grid[pos] = int(pnum)
            # add random walls if needed
            randwalls = self.game_params.get('random_walls', 0)
            while randwalls>0:
                empty = list(self.enumerate_coords_with_value(0))
                self.grid[random.choice(empty)] = -1
                randwalls -= 1
    
    # returns an iterable of ( (move_positions), game_after_move )
    # subclasses must override
    def successors(self):
        if 0: yield None
        
    def successors_for_ai(self):
        return self.successors()
        
    # cached_successors is not called by AI, only by UI controller to determine which selected moves are legal
    def cached_successors(self):
        if not hasattr(self, '_cached_successors'):
            self._cached_successors = dict( (s[0],s[1]) for s in self.successors())
        return self._cached_successors
        
    # methods to determine valid partial/total moves
    def valid_partial_move(self, move):
        return tuple(move) in (s[:len(move)] for s in self.cached_successors())
        
    def valid_move(self, move):
        return tuple(move) in self.cached_successors()
        
    def next_state_for_move(self, move):
        return self.cached_successors().get(tuple(move), None)
        
    def has_scores(self):
        return True
        
    # default implementation counts grid elements with value=pnum
    def player_score(self, pnum):
        return sum(1 for v in self.itervalues() if v==pnum)
            
    def winning_player(self):
        s1 = self.player_score(1)
        s2 = self.player_score(2)
        if s1>s2: return 1
        if s2>s1: return 2
        return 0
        
    def animations_for_move(self, move, newstate):
        """ Returns a PGPieceAnimation.AnimationGroup representing animations that correspond
            to this move. This implementation creates animations for all changing cells; subclasses can override.
        """
        import PGPieceAnimation
        animations = []
        for c in self.itercoords():
            v0, v1 = self.grid[c], newstate.grid[c]
            if v0!=v1:
                segment = PGPieceAnimation.AnimationSegment(from_position=c, to_position=c, from_value=v0, to_value=v1,
                                                            start_time=0.0, end_time=PGPieceAnimation.DEFAULT_DURATION)
                # for "new" and "remove" animations, keep to and from values the same, only scale changes
                if v0==0:
                    segment.action = PGPieceAnimation.NEW
                    segment.animation = PGPieceAnimation.GROW
                    segment.from_value = v1
                elif v1==0:
                    segment.action = PGPieceAnimation.REMOVE
                    segment.animation = PGPieceAnimation.GROW
                    segment.to_value = v0
                else:
                    segment.action = PGPieceAnimation.CHANGE
                    segment.animation = PGPieceAnimation.FADE
                   
                position = None if v0==0 else c
                animations.append(PGPieceAnimation.Animation(position, [segment]))
                
        return PGPieceAnimation.AnimationGroup(animations)
        
    # subclasses should override to read additional state from serialized game
    def finish_deserialization(self, obj):
        pass
        
        
def game_state_class(gname):
    # look for explicit STATE_CLASS attribute, then [Game name]GameState
    gmodule = __import__(gname)
    if hasattr(gmodule, 'STATE_CLASS'): return gmodule.STATE_CLASS
    return getattr(gmodule, gname+'GameState')
    
class Game:
    def __init__(self, gamename, params, initstate=True):
        self.game_name = gamename
        self.game_params = params
        self.game_states = []
        self.current_state_index = -1
        if initstate:
            state = game_state_class(gamename)(None, params)
            self.push_game_state(state)
            
    def push_game_state(self, state):
        # clear out any previously undone moves
        if self.current_state_index < -1:
            del self.game_states[len(self.game_states)+self.current_state_index+1:]
        self.game_states.append(state)
        self.current_state_index = -1
        
    def can_undo_move(self):
        return -self.current_state_index < len(self.game_states)
        
    def undo_move(self):
        if not self.can_undo_move():
            raise RuntimeError("Can't undo")
        self.current_state_index -= 1
        
    def can_redo_move(self):
        return self.current_state_index < -1
        
    def redo_move(self):
        if not self.can_redo_move():
            raise RuntimeError("Can't redo")
        self.current_state_index += 1
        
    def has_started(self):
        # returns true if at least one move has been made
        return len(self.game_states)>1
        
    def move_count(self):
        """ Returns the number of moves made during this game (0 for a just-started game) """
        return len(self.game_states) + self.current_state_index
        
    current_state = property(lambda self: self.game_states[self.current_state_index], None)
        
    def make_move(self, move):
        if self.current_state.is_game_over(): raise RuntimeError("Game is over")
        nextstate = self.current_state.next_state_for_move(move)
        if nextstate is None: raise ValueError("Invalid move"+str(move))
        self.push_game_state(nextstate)
        return self.current_state
        
    def make_ai_move(self, ai):
        if self.current_state.is_game_over(): raise RuntimeError("Game is over")
        move, nextstate, nmoves, t = ai.best_move(self.current_state)
        self.push_game_state(nextstate)
        return (move, self.current_state, nmoves, t)
        
    def serialize(self):
        return {'game':self.game_name, 'params': self.game_params, 'current_state': self.current_state_index,
                'game_states': [s.serialize() for s in self.game_states]}
                
    def deserialize_game_state(self, obj):
        state = game_state_class(self.game_name)(None, self.game_params, reset=False)
        state.num_players = obj['num_players']
        state.current_player = obj['current_player']
        state.grid = numpy.array(obj['values']).reshape(obj['shape'])
        # semi-hack: extra_state is likely to contain a sequence of scores, which must be a list and not a tuple so they can be assigned
        state.extra_state = PGUtils.tuples_to_lists(obj.get('extra_state'))
        state.finish_deserialization(obj)
        return state
        
        
def deserialize_game(obj):
    params = obj.get('params', {})
    gname = obj['game']
    game = Game(gname, params, initstate=False)
    game.game_states = [game.deserialize_game_state(s) for s in obj['game_states']]
    game.current_state_index = obj['current_state']
    return game
    
    
def testa():
    params = {'dimensions':(7,7), 'positions':{(0,0):1, (0,6):2, (6,6):1, (6,0):2}}
    game = Game('Ataxx', params)
    game.set_ai(2, GenericAI({}))
    
    assert game.current_state.grid.shape==(7,7)
    assert game.current_state.grid[0,0]==1
    assert game.current_state.grid[(0,6)]==2
    
    assert game.current_state.current_player==1
    assert game.current_state.next_player()==2
    
    assert game.current_state.player_score(1)==game.current_state.player_score(2)==2
    
    assert not game.is_ai_turn()
    
    return game
    
def testr():
    from GenericAI import GenericAI
    params = {'dimensions':(8,8), 'positions':{(3,3):1, (3,4):2, (4,4):1, (4,3):2}}
    game = Game("Reversi", params)
    game.set_ai(2, GenericAI({}))
    return game
    
def testta():
    from GenericAI import GenericAI
    params = {'dimensions':(8,8)}
    game = Game("ThinkAhead", params)
    game.set_ai(2, GenericAI({}))
    return game
    